# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Device",
    "device_set",
    ]

from storm.locals import (
    AutoReload,
    Int,
    Reference,
    Store,
    )

from zope.cachedescriptors.property import Lazy
from zope.component import getUtility
from zope.interface import implements

from lazr.restful.interfaces import IServiceRootResource

from lpresults.repository.formats.pnp import pnp_vendor_to_int

from lpresults.hardware.interfaces.device import (
    IDevice,
    IDeviceSet,
    )
from lpresults.hardware.models.devicebus import DeviceBus
from lpresults.hardware.models.devicecategory import DeviceCategory
from lpresults.hardware.models.deviceproduct import DeviceProduct
from lpresults.hardware.models.devicesubproduct import DeviceSubproduct
from lpresults.hardware.models.devicevendor import DeviceVendor
from lpresults.hardware.models.store import get_hardware_store


class Device:

    __storm_table__ = "device"

    implements(IDevice)

    id = Int(primary=True, allow_none=False, default=AutoReload)
    bus_id = Int(name="bus", allow_none=False)
    product_id = Int(name="product", allow_none=False)
    # Variant attributes:
    vendor_id = Int(name="vendor")
    subproduct_id = Int(name="subproduct")
    subvendor_id = Int(name="subvendor")
    category_id = Int(name="category")

    bus = Reference(bus_id, DeviceBus.id)
    vendor = Reference(
        (bus_id, vendor_id),
        (DeviceVendor.bus_id, DeviceVendor.vendor_id))
    subproduct = Reference(
        (product_id, vendor_id, subproduct_id, subvendor_id),
        (DeviceSubproduct.product_id, DeviceSubproduct.vendor_id,
            DeviceSubproduct.subproduct_id, DeviceSubproduct.subvendor_id))
    subvendor = Reference(
        (bus_id, subvendor_id),
        (DeviceVendor.bus_id, DeviceVendor.vendor_id))
    category = Reference(category_id, DeviceCategory.id)

    def __init__(
        self, bus, product_id, vendor_id=None, subproduct_id=None,
        subvendor_id=None, category=None):

        self.bus = bus
        self.product_id = product_id
        self.vendor_id = vendor_id
        self.subproduct_id = subproduct_id
        self.subvendor_id = subvendor_id
        self.category = category

    @property
    def __parent__(self):
        return getUtility(IServiceRootResource)

    @property
    def __path_parts__(self):
        return ("devices", self.path,)

    @Lazy
    def product(self):
        # XXX Using find instead of a reference because the vendor_id
        #     might be null.
        return Store.of(self).find(
            DeviceProduct, bus_id=self.bus_id, product_id=self.product_id,
            vendor_id=self.vendor_id).one()

    @property
    def path(self):
        path = "%s:%.4X" % (self.bus_name, self.product_id or 0)
        if self.vendor_id is not None:
            path += ":%.4X" % self.vendor_id
            if self.subproduct_id is not None:
                path += ":%.4X" % self.subproduct_id

            if self.subvendor_id is not None:
                path += ":%.4X" % self.subvendor_id

        if self.category_name:
            path += "-%s" % self.category_name

        return path

    @property
    def bus_name(self):
        """See `IHasDeviceBus`."""
        return self.bus.name

    @property
    def category_name(self):
        """See `IHasDeviceCategory`."""
        return self.category.name if self.category else None

    @property
    def product_name(self):
        """See `IHasDeviceProduct`."""
        return self.product.name if self.product else None

    @property
    def vendor_name(self):
        """See `IHasDeviceVendor`."""
        return self.vendor.name if self.vendor else None


class DeviceSet:

    implements(IDeviceSet)

    def get(self, id):
        """See `IDeviceSet`."""
        return get_hardware_store().get(Device, id)

    def getByName(
        self, bus_name, category_name=None, product_name=None,
        vendor_name=None, product_id=None, vendor_id=None,
        subproduct_id=None, subvendor_id=None):
        """See `IDeviceSet`."""
        # XXX Special case for PNP where the vendor_id can be generated.
        if bus_name == "pnp":
            vendor_id = pnp_vendor_to_int(vendor_name)
        id = get_hardware_store().execute(
            "SELECT get_device(?, ?, ?, ?, ?, ?, ?, ?)",
            (bus_name, category_name, product_name, vendor_name,
            product_id, vendor_id, subproduct_id, subvendor_id,)).get_one()[0]
        return self.get(id)


device_set = DeviceSet()
