# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "DeviceState",
    "DeviceStateTargetMixin",
    "HasDeviceStatesMixin",
    ]

from storm.expr import LeftJoin
from storm.locals import (
    And,
    AutoReload,
    Int,
    Join,
    Reference,
    Store,
    )

from zope.component import getUtility
from zope.interface import implements

from lpresults.database.constants import (
    INFINITY,
    UTC_NOW,
    )
from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTime
from lpresults.database.search import (
    SearchGreaterThan,
    SearchLessThanOrEqualTo,
    search_to_expression,
    )

from lpresults.registry.models.search import SearchSet

from lpresults.hardware.interfaces.device import (
    IDevice,
    IDeviceSet,
    )
from lpresults.hardware.interfaces.devicestate import (
    IDeviceState,
    IDeviceStateSet,
    DeviceStateSearchParams,
    )
from lpresults.hardware.models.device import Device
from lpresults.hardware.models.devicebus import DeviceBus
from lpresults.hardware.models.devicecategory import DeviceCategory
from lpresults.hardware.models.devicedriver import DeviceDriver


class DeviceState:

    __storm_table__ = "devicestate"

    implements(IDeviceState)
    delegates(IDevice, context="device")

    id = Int(primary=True, allow_none=False, default=AutoReload)
    system_unit_id = Int(name="system_unit", allow_none=False)
    device_id = Int(name="device", allow_none=False)
    driver_id = Int(name="driver")
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)
    date_deleted = UTCDateTime(allow_none=False, default=INFINITY)

    device = Reference(device_id, Device.id)
    driver = Reference(driver_id, DeviceDriver.id)

    def __init__(
        self, system_unit, device, driver=None, date_created=None):
        """Initialize a processor state."""
        self.system_unit = system_unit
        self.device = device
        self.driver = driver

        if date_created is not None:
            self.date_created = date_created

    @property
    def __parent__(self):
        return self.device

    @property
    def __url_path__(self):
        return self.driver_name or ""

    @property
    def driver_name(self):
        return self.driver.name if self.driver else None


class DeviceStateSet(SearchSet):
    """See `IDeviceStateSet`."""

    implements(IDeviceStateSet)

    search_params_factory = DeviceStateSearchParams

    def __init__(self, target):
        self.target = target

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(DeviceStateSet, self).buildSearchTables(
            search_params)

        tables.append(Join(
            Device,
            Device.id == DeviceState.device_id))
        tables.append(Join(
            DeviceBus,
            DeviceBus.id == Device.bus_id))
        tables.append(LeftJoin(
            DeviceCategory,
            DeviceCategory.id == Device.category_id))
        tables.append(LeftJoin(
            DeviceDriver,
            DeviceDriver.id == DeviceState.driver_id))

        return tables

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(DeviceStateSet, self).buildSearchExpressions(
            search_params)

        expressions.append(DeviceState.system_unit == self.target)

        if search_params.date_created is not None:
            expressions.append(And(
                search_to_expression(
                    DeviceState.date_created,
                    SearchLessThanOrEqualTo(search_params.date_created)),
                search_to_expression(
                    DeviceState.date_created,
                    SearchGreaterThan(search_params.date_created))))

        if search_params.created_before is not None:
            expressions.append(
                search_to_expression(
                    DeviceState.date_created,
                    SearchLessThanOrEqualTo(search_params.created_before)))

        if search_params.created_after is not None:
            expressions.append(
                search_to_expression(
                    DeviceState.date_created,
                    SearchGreaterThan(search_params.created_after)))

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return (DeviceState, Device, DeviceBus, DeviceCategory, DeviceDriver,)

    def getSearchStore(self, search_params):
        return Store.of(self.target)


class HasDeviceStatesMixin:

    def getDeviceStates(
        self, date_created=None, created_before=None, created_after=None):
        """See `IDeviceStateTarget`."""
        params = DeviceStateSearchParams(
            date_created=date_created,
            created_before=created_before,
            created_after=created_after)
        return DeviceStateSet(self).search(params)


class DeviceStateTargetMixin(HasDeviceStatesMixin):

    def addDeviceState(
        self, bus_name, category_name=None, product_name=None,
        vendor_name=None, product_id=None, vendor_id=None,
        subproduct_id=None, subvendor_id=None, driver_name=None,
        date_created=None):
        """See `IDeviceStateTarget`."""
        device = getUtility(IDeviceSet).getByName(
            bus_name, category_name, product_name, vendor_name,
            product_id, vendor_id, subproduct_id, subvendor_id)

        store = Store.of(self)
        if driver_name:
            driver = store.find(DeviceDriver, name=driver_name).one()
            if not driver:
                driver = DeviceDriver(driver_name)
        else:
            driver = None

        state = store.find(
            DeviceState, system_unit=self, device=device, driver=driver).one()
        if not state:
            state = DeviceState(self, device, driver, date_created)
            store.add(state)

        return state
