# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "HasProcessorStatesMixin",
    "ProcessorState",
    "ProcessorStateTargetMixin",
    ]

from storm.locals import (
    And,
    AutoReload,
    Int,
    Join,
    Reference,
    Store,
    )

from zope.component import getUtility
from zope.interface import implements

from lazr.restful.simple import TraverseWithGet

from lpresults.database.constants import (
    INFINITY,
    UTC_NOW,
    )
from lazr.delegates import delegates

from lpresults.database.datetime import UTCDateTime

from lpresults.registry.models.search import SearchSet

from lpresults.hardware.interfaces.processor import (
    IProcessor,
    IProcessorSet,
    )
from lpresults.hardware.interfaces.processorstate import (
    IProcessorState,
    IProcessorStateSet,
    ProcessorStateSearchParams,
    )
from lpresults.hardware.models.processor import Processor
from lpresults.hardware.models.processorplatform import ProcessorPlatform


class ProcessorState:

    __storm_table__ = "processorstate"

    __url_path__ = "processor"

    implements(IProcessorState)
    delegates(IProcessor, context="processor")

    id = Int(primary=True, allow_none=False, default=AutoReload)
    system_unit_id = Int(name="system_unit", allow_none=False)
    processor_id = Int(name="processor", allow_none=False)
    bogomips = Int(allow_none=False)
    speed = Int(allow_none=False)
    count = Int(allow_none=False)
    date_created = UTCDateTime(allow_none=False, default=UTC_NOW)
    date_deleted = UTCDateTime(allow_none=False, default=INFINITY)

    processor = Reference(processor_id, Processor.id)

    def __init__(
        self, system_unit, processor, bogomips, speed, count,
        date_created=None):
        """Initialize a processor state."""
        self.system_unit = system_unit
        self.processor = processor
        self.bogomips = bogomips
        self.speed = speed
        self.count = count

        if date_created is not None:
            self.date_created = date_created

    @property
    def __parent__(self):
        return self.system_unit


class ProcessorStateSet(SearchSet):
    """See `IProcessorStateSet`."""

    implements(IProcessorStateSet)

    search_params_factory = ProcessorStateSearchParams

    def __init__(self, target):
        self.target = target

    def buildSearchTables(self, search_params):
        """See SearchSet."""
        tables = super(ProcessorStateSet, self).buildSearchTables(
            search_params)

        tables.append(
            Join(
                Processor,
                Processor.id == ProcessorState.processor_id))
        tables.append(
            Join(
                ProcessorPlatform,
                ProcessorPlatform.id == Processor.platform_id))

        return tables

    def buildSearchExpressions(self, search_params):
        """See SearchSet."""
        expressions = super(ProcessorStateSet, self).buildSearchExpressions(
            search_params)

        expressions.append(ProcessorState.system_unit == self.target)

        if search_params.date_created is not None:
            expressions.append(And(
                ProcessorState.date_created <= search_params.date_created,
                ProcessorState.date_created > search_params.date_created))

        if search_params.created_before is not None:
            expressions.append(
                ProcessorState.date_created <= search_params.created_before)

        if search_params.created_after is not None:
            expressions.append(
                ProcessorState.date_created > search_params.created_after)

        return expressions

    def getSearchResultTables(self):
        """See SearchSet."""
        return (ProcessorState, Processor, ProcessorPlatform,)

    def getSearchStore(self, search_params):
        return Store.of(self.target)


class HasProcessorStatesMixin(TraverseWithGet):

    def get(self, request, name):
        if name == "processor":
            return self.getProcessorStates().any()

        return super(HasProcessorStatesMixin, self).get(request, name)

    def getProcessorStates(
        self, date_created=None, created_before=None, created_after=None):
        """See `IProcessorStateTarget`."""
        params = ProcessorStateSearchParams(
            date_created=date_created,
            created_before=created_before,
            created_after=created_after)
        target = self.getStateTarget()
        return ProcessorStateSet(target).search(params)

    def getStateTarget(self):
        return self


class ProcessorStateTargetMixin(HasProcessorStatesMixin):

    def setProcessorState(
        self, platform_name, make, model, model_number, model_version,
        model_revision, cache, other, bogomips, speed, count,
        date_created=None):
        """See `IProcessorStateTarget`."""
        processor = getUtility(IProcessorSet).getByName(
            platform_name, make, model, model_number, model_version,
            model_revision, cache, other)

        target = self.getStateTarget()
        store = Store.of(target)
        state = store.find(
            ProcessorState, system_unit=target, processor=processor).one()
        if not state:
            state = ProcessorState(
                target, processor, bogomips, speed, count, date_created)
            store.add(state)

        return state
