# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "Application",
    "ApplicationError",
    ]

import sys

from httplib import BadStatusLine
from optparse import (
    OptionGroup,
    OptionParser,
    OptionValueError,
    )
from urllib2 import (
    HTTPError,
    URLError,
    urlopen,
    )

from lpresults.scripts.logger import get_logger
from lpresults.scripts.retry import retry_except


class ApplicationError(Exception):
    """Error raised when application fails."""

    status = 1


class Application:

    # Applicaiton defaults
    usage = "Usage: %prog [OPTIONS]"

    # Logging defaults
    default_log_level = "info"

    def __init__(self, name=None):
        """Construct new Application.

        :param name: A short name for the application.
        """
        if name is None:
            self._name = self.__class__.__name__.lower()
        else:
            self._name = name

    @property
    def name(self):
        """Enable subclasses to override with command-line arguments."""
        return unicode(self._name)

    @retry_except(ApplicationError)
    def getURL(self, url):
        """Helper to get a URL and retry on recoverable errors.

        :param url: URL to get.
        """
        try:
            return urlopen(url)
        except (URLError, BadStatusLine), error:
            # 404 errors indicate a data inconsistency: more than one
            # try to open the URL is pointless.
            #
            # Note that URLError is a base class of HTTPError.
            if isinstance(error, HTTPError) and error.code == 404:
                raise

            # HTTPErrors with a 5xx error code ("server problem") are
            # a reason to retry the access again, as well as generic,
            # non-HTTP, URLErrors like "connection refused" or
            # BadStatusLine like "connection closed".
            if (isinstance(error, HTTPError)
                and 500 <= error.code <= 599
                or isinstance(error, URLError) and
                    not isinstance(error, HTTPError)
                or isinstance(error, BadStatusLine)):
                raise ApplicationError(str(error))

            else:
                raise

    def addOptions(self, parser):
        group = OptionGroup(parser, "Logging options")
        group.add_option("-l", "--log",
            metavar="FILE",
            help="The file to write the log to.")
        group.add_option("--log-level",
            default=self.default_log_level,
            help="One of debug, info, warning, error or critical.")
        parser.add_option_group(group)

    def parseOptions(self, options, args):
        log = options.log
        if log:
            if log == "-":
                log_stream, log_file = sys.stdout, None
            else:
                log_stream, log_file = None, log
        else:
            log_stream, log_file = sys.stderr, None
        self.logger = get_logger(
            options.log_level, log_stream, log_file)

    def process(self):
        """Process the application here. Must be defined."""
        raise NotImplementedError()

    def preProcess(self):
        """
        This should set up any state necessary before loading and
        processing the Application.
        """

    def postProcess(self, error):
        """This will be called after the processing has finished."""

    def run(self, args=None):
        """Run the application."""
        parser = OptionParser(usage=self.usage)
        self.addOptions(parser)
        options, args = parser.parse_args(args)
        try:
            self.parseOptions(options, args)
        except OptionValueError, error:
            parser.error(error)

        error = None
        self.preProcess()
        try:
            self.process()
        except ApplicationError, error:
            self.logger.error(str(error))
            sys.exit(error.status)
        except Exception:
            self.logger.exception("Unhandled exception")
            sys.exit(1)
        finally:
            self.postProcess(error)
