# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""
Added support for inserting test runs backwards in time.
"""

statements = [
"""DROP FUNCTION next_distroseries_sequence(INTEGER)""",
"""DROP FUNCTION next_projectseries_sequence(INTEGER)""",
"""DROP TRIGGER testresult_delete_after_trigger ON testresult""",
"""DROP FUNCTION testresult_delete_after()""",
"""
CREATE OR REPLACE FUNCTION testrun_insert_before() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
BEGIN
    IF NEW.distro_series IS NOT NULL THEN
        UPDATE distroseriestwin SET counter = counter + 1
            WHERE id = NEW.distro_series RETURNING counter INTO NEW.sequence;
    ELSIF NEW.project_series IS NOT NULL THEN
        UPDATE projectseriestwin SET counter = counter + 1
            WHERE id = NEW.project_series RETURNING counter INTO NEW.sequence;
    END IF;

    RETURN NEW;
END
$$
""",
"""
CREATE OR REPLACE FUNCTION testrun_insert_after() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
BEGIN
    INSERT INTO testresult (
            test_case,
            status,
            output,
            date_created,
            date_deleted)
        SELECT
                testresult.test_case,
                testresult.status,
                testresult.output,
                MIN(testrun.date_created),
                testresult.date_deleted
            FROM testresult
            JOIN testrun
                ON testrun.date_created > NEW.date_created
                AND testrun.date_created < testresult.date_deleted
            WHERE testresult.date_created < NEW.date_created
            AND CASE WHEN NEW.distro_series IS NOT NULL THEN
                    testrun.distro_series = NEW.distro_series
                ELSE
                    testrun.project_series = NEW.project_series
                END
            GROUP BY
                testresult.test_case,
                testresult.status,
                testresult.output,
                testresult.date_deleted;

    UPDATE testresult
        SET date_deleted = NEW.date_created
        FROM testcase
        WHERE testresult.date_created < NEW.date_created
        AND testresult.date_deleted > NEW.date_created
        AND testresult.test_case = testcase.id
        AND CASE WHEN NEW.distro_series IS NOT NULL THEN
                testcase.distro_series = NEW.distro_series
            ELSE
                testcase.project_series = NEW.project_series
            END;

    RETURN NEW;
END
$$
""",
"""
DROP TRIGGER testrun_insert_after_trigger
    ON testrun
""",
"""
CREATE TRIGGER testrun_insert_after_trigger
    AFTER INSERT ON testrun
    FOR EACH ROW
    EXECUTE PROCEDURE testrun_insert_after()
""",
"""
CREATE FUNCTION testrun_update_before() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
BEGIN
    IF OLD.distro_series <> NEW.distro_series THEN
        RAISE EXCEPTION 'Cannot update distro_series';
    ELSIF OLD.project_series <> NEW.project_series THEN
        RAISE EXCEPTION 'Cannot update project_series';
    ELSIF OLD.date_created <> NEW.date_created THEN
        RAISE EXCEPTION 'Cannot update project_series';
    ELSE
        RETURN NEW;
    END IF;
END;
$$
""",
"""
GRANT EXECUTE
    ON FUNCTION testrun_update_before()
    TO lpresults
""",
"""
CREATE TRIGGER testrun_update_before_trigger
    BEFORE UPDATE ON testrun
    FOR EACH ROW
    EXECUTE PROCEDURE testrun_update_before()
""",
"""
CREATE FUNCTION testrun_delete_after() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
DECLARE
    next_testrun RECORD;
BEGIN
    SELECT
        INTO next_testrun
            testrun.date_created
        FROM testrun
        WHERE testrun.date_created > OLD.date_created
        AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                testrun.distro_series = OLD.distro_series
            ELSE
                testrun.project_series = OLD.project_series
            END
        ORDER BY testrun.date_created ASC LIMIT 1;

    -- Between
    IF next_testrun.date_created IS NOT NULL THEN
        UPDATE testresult tr1
            SET date_deleted = tr2.date_deleted
            FROM testresult tr2, testcase tc
            WHERE tr1.test_case = tr2.test_case
            AND tr1.status = tr2.status
            AND decode(md5(tr1.output), 'hex') = decode(md5(tr2.output), 'hex')
            AND tr1.date_deleted = OLD.date_created
            AND tr2.date_created = next_testrun.date_created
            AND tr1.test_case = tc.id
            AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                    tc.distro_series = OLD.distro_series
                ELSE
                    tc.project_series = OLD.project_series
                END;
        IF FOUND THEN
            DELETE FROM testresult tr2
                USING testresult tr1, testcase tc
                WHERE (
                    tr1.test_case = tr2.test_case
                    AND tr1.status = tr2.status
                    AND decode(md5(tr1.output), 'hex')
                        = decode(md5(tr2.output), 'hex')
                    AND tr2.date_created = next_testrun.date_created)
                OR (
                    tr2.date_created = OLD.date_created
                    AND tr2.date_deleted = next_testrun.date_created)
                AND tr1.test_case = tc.id
                AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                        tc.distro_series = OLD.distro_series
                    ELSE
                        tc.project_series = OLD.project_series
                    END;
        ELSE
            DELETE FROM testresult tr
                USING testcase tc
                WHERE tr.date_created = OLD.date_created
                AND tr.date_deleted = next_testrun.date_created
                AND tr.test_case = tc.id
                AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                        tc.distro_series = OLD.distro_series
                    ELSE
                        tc.project_series = OLD.project_series
                    END;
        END IF;

        -- After
        UPDATE testresult tr
            SET date_created = next_testrun.date_created
            FROM testcase
            WHERE tr.date_created = OLD.date_created
            AND tr.test_case = testcase.id
            AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                    testcase.distro_series = OLD.distro_series
                ELSE
                    testcase.project_series = OLD.project_series
                END;
    ELSE
        next_testrun.date_created = '3000-01-01';
        DELETE FROM testresult tr
            USING testcase tc
            WHERE tr.date_created = OLD.date_created
            AND tr.test_case = tc.id
            AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                    tc.distro_series = OLD.distro_series
                ELSE
                    tc.project_series = OLD.project_series
                END;
    END IF;

    -- Before
    UPDATE testresult tr
        SET date_deleted = next_testrun.date_created
        FROM testcase tc
        WHERE tr.date_deleted = OLD.date_created
        AND tr.test_case = tc.id
        AND CASE WHEN OLD.distro_series IS NOT NULL THEN
                tc.distro_series = OLD.distro_series
            ELSE
                tc.project_series = OLD.project_series
            END;

    RETURN OLD;
END;
$$
""",
"""
GRANT EXECUTE
    ON FUNCTION testrun_delete_after()
    TO lpresults
""",
"""
CREATE TRIGGER testrun_delete_after_trigger
    AFTER DELETE ON testrun
    FOR EACH ROW
    EXECUTE PROCEDURE testrun_delete_after()
""",
"""
CREATE OR REPLACE FUNCTION testresult_insert_before() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
DECLARE
    next_date_created DATE;
    next_testrun DATE;
    previous_date_deleted DATE;
BEGIN
    IF NEW.date_deleted < '3000-01-01' THEN
        RETURN NEW;
    END IF;

    SELECT INTO next_testrun
            CASE WHEN MIN(date_created) > NEW.date_created
            THEN MIN(date_created)
            ELSE MIN(date_deleted)
            END
        FROM testresult tr
        JOIN testcase tc1
            ON tr.test_case = tc1.id
        JOIN testcase tc2
            ON tc1.distro_series IS NOT DISTINCT FROM tc2.distro_series
            AND tc1.project_series IS NOT DISTINCT FROM tc2.project_series
        WHERE date_deleted > NEW.date_created
        AND tc2.id = NEW.test_case;
    IF next_testrun < '3000-01-01' THEN
        -- Set date created of next test result.
        UPDATE testresult
            SET date_created = CASE
                WHEN testresult.status = NEW.status
                    AND testresult.output = NEW.output
                    AND testresult.date_created <> NEW.date_created
                THEN NEW.date_created
                ELSE next_testrun
                END
            WHERE testresult.test_case = NEW.test_case
            AND testresult.date_created <= next_testrun
            AND testresult.date_created >= NEW.date_created
            RETURNING date_created INTO next_date_created;
        IF next_date_created != NEW.date_created THEN
            next_date_created := next_testrun;
            NEW.date_deleted := next_testrun;
        END IF;
    ELSE
        next_date_created := '3000-01-01';
    END IF;

    -- Set date deleted of previous test result.
    UPDATE testresult
        SET date_deleted = CASE
            WHEN testresult.status = NEW.status
                AND testresult.output = NEW.output
            THEN next_date_created
            ELSE NEW.date_created
            END
        WHERE testresult.test_case = NEW.test_case
        AND testresult.date_created <= NEW.date_created
        AND testresult.date_deleted >= NEW.date_created
        RETURNING date_deleted INTO previous_date_deleted;
    IF previous_date_deleted = next_date_created THEN
        IF next_date_created = NEW.date_created THEN
            DELETE FROM testresult
                WHERE test_case = NEW.test_case
                AND testresult.date_created = next_testrun;
        END IF;
        RETURN NULL;
    ELSIF next_date_created = NEW.date_created
          AND previous_date_deleted IS NOT NULL THEN
        RETURN NULL;
    END IF;

    RETURN NEW;
END;
$$
""",
"""
CREATE OR REPLACE FUNCTION testresult_update_before() RETURNS TRIGGER
LANGUAGE plpgsql AS
$$
BEGIN
    IF OLD.test_case <> NEW.test_case THEN
        RAISE EXCEPTION 'Cannot update test_case';
    ELSIF OLD.status <> NEW.status THEN
        RAISE EXCEPTION 'Cannot update status';
    ELSIF OLD.output <> NEW.output THEN
        RAISE EXCEPTION 'Cannot update output';
    ELSIF NEW.date_deleted = NEW.date_created THEN
        DELETE FROM testresult
            WHERE test_case = OLD.test_case
            AND date_created = OLD.date_created;
        RETURN NULL;
    ELSE
        RETURN NEW;
    END IF;
END;
$$
""",
]


def apply(store):
    print __doc__
    for statement in statements:
        store.execute(statement)
