# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

from datetime import datetime

from unittest import defaultTestLoader

from lpresults.database.store import get_store

from lpresults.tracker.testing.cases import TestCaseWithFactory
from lpresults.tracker.testing.layers import DatabaseLayer


class TestResultTestCase(TestCaseWithFactory):

    layer = DatabaseLayer

    default_status = 0
    default_output = u""
    default_date_created = datetime(2010, 1, 1)

    def setUp(self):
        super(TestResultTestCase, self).setUp()

        self.store = get_store("test-1")

        self.project_series = self.factory.getUniqueInteger()
        self.project = self.factory.getUniqueInteger()
        self.store.execute(
            "INSERT INTO projectseriestwin VALUES (?, ?)",
            (self.project_series, self.project),
            noresult=True,
            )

        name = self.factory.getUniqueString(u"test-name")
        self.test_case = self.store.execute(
            "SELECT get_test_case(?, ?, ?)",
            (None, self.project_series, name,),
            ).get_one()[0]

    def tearDown(self):
        super(TestResultTestCase, self).tearDown()

        self.store.execute(
            "DELETE FROM projectseriestwin WHERE id = ?",
            (self.project_series,),
            )

    def getTestResultQuery(
        self, status=None, output=None, date_created=None, date_deleted=None):
        query = ["test_case = %d" % self.test_case]
        if status is not None:
            query.append("status = %d" % status)
        if output is not None:
            query.append("output = '%s'" % output)
        if date_created is not None:
            query.append("date_created = '%s'" % date_created)
        if date_deleted is not None:
            query.append("date_deleted = '%s'" % date_deleted)

        return " AND ".join(query)

    def insertTestResult(self, status=None, output=None, date_created=None):
        if status is None:
            status = self.default_status
        if output is None:
            output = self.default_output
        if date_created is None:
            date_created = self.default_date_created

        self.store.execute(
            "INSERT INTO testresult VALUES (?, ?, ?, ?)",
            (self.test_case, status, output, date_created,),
            noresult=True,
            )

    def removeTestResult(self, **kwargs):
        query = self.getTestResultQuery(**kwargs)
        self.store.execute("DELETE FROM testresult WHERE %s" % query)

    def countTestResults(self, **kwargs):
        query = self.getTestResultQuery(**kwargs)
        return self.store.execute(
            "SELECT COUNT(*) FROM testresult WHERE %s" % query).get_one()[0]

    def getTestResult(self, **kwargs):
        keys = [
            "test_case", "status", "output", "date_created", "date_deleted"]
        select_expression = ", ".join(keys)
        query = self.getTestResultQuery(**kwargs)
        values = self.store.execute(
            "SELECT %s FROM testresult WHERE %s ORDER BY date_created"
            % (select_expression, query)).get_all()
        if len(values) != 1:
            self.fail("One test result expected, %d found" % len(values))

        return dict(zip(keys, values[0]))

    def assertCountTestResults(self, count, **kwargs):
        test_result_count = self.countTestResults(**kwargs)
        self.assertEquals(test_result_count, count)


class TestTestResultInsert(TestResultTestCase):

    def test_insertOneResult(self):
        self.assertCountTestResults(0)
        self.insertTestResult()
        self.assertCountTestResults(1)

        # Only the first test result should exist
        test_result = self.getTestResult()
        self.assertEquals(test_result["test_case"], self.test_case)
        self.assertEquals(test_result["status"], self.default_status)
        self.assertEquals(test_result["output"], self.default_output)
        self.assertEquals(
            test_result["date_created"], self.default_date_created)
        self.assertTrue(
            test_result["date_deleted"] > self.default_date_created)

    def test_insertTwoDifferentResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # There should be two rows after creating different test results
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=1, date_created=second_date)
        self.assertCountTestResults(2)

        # Both test results should exist
        first_result = self.getTestResult(date_created=first_date)
        self.assertNotEquals(first_result, None)
        second_result = self.getTestResult(date_created=second_date)
        self.assertNotEquals(second_result, None)

        # The date_deleted of the first test result
        # should be the same as the date_created of the second test result
        self.assertEquals(
            first_result["date_deleted"],
            second_result["date_created"])

    def test_insertTwoSimilarResults(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # There should be two rows after creating different test results
        self.insertTestResult(status=0, date_created=first_date)
        self.insertTestResult(status=0, date_created=second_date)

        # Only the first test result should exist
        test_result = self.getTestResult(date_created=first_date)
        self.assertEquals(test_result["status"], 0)
        self.assertEquals(test_result["date_created"], first_date)

        self.assertCountTestResults(0, date_created=second_date)

    def test_insertTwoSimilarResultsReversed(self):
        first_date = datetime(2010, 1, 1)
        self.assertCountTestResults(0, date_created=first_date)
        second_date = datetime(2010, 1, 2)
        self.assertCountTestResults(0, date_created=second_date)

        # Add test results in reversed order
        self.insertTestResult(status=0, date_created=second_date)
        self.insertTestResult(status=0, date_created=first_date)

        # Only the first test result should exist
        test_result = self.getTestResult(date_created=first_date)
        self.assertEquals(test_result["status"], 0)
        self.assertEquals(test_result["date_created"], first_date)

        self.assertCountTestResults(0, date_created=second_date)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
