# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = []

from datetime import timedelta

from unittest import defaultTestLoader

from lpresults.database.constants import INFINITY_DATETIME

from lpresults.tracker.schema.test.tests.test_testresult import (
    TestResultTestCase,
    )


class TestRunTestCase(TestResultTestCase):

    def insertTestResult(self, date_created=None, **kwargs):
        self.insertTestRun(date_created=date_created)
        super(TestRunTestCase, self).insertTestResult(
            date_created=date_created, **kwargs)

    def getTestRunQuery(self, date_created=None):
        query = ["project_series = %d" % self.project_series]
        if date_created is not None:
            query.append("date_created = '%s'" % date_created)

        return " AND ".join(query)

    def insertTestRun(self, date_created=None):
        if date_created is None:
            date_created = self.default_date_created

        self.store.execute(
            "INSERT INTO testrun (project_series, date_created)"
            " VALUES (?, ?)",
            (self.project_series, date_created,),
            noresult=True,
            )

    def deleteTestRun(self, date_created=None):
        if date_created is None:
            date_created = self.default_date_created

        self.store.execute(
            "DELETE FROM testrun"
            " WHERE project_series = ? AND date_created = ?",
            (self.project_series, date_created,),
            noresult=True,
            )

    def getTestRun(self, **kwargs):
        keys = ["project_series", "date_created"]
        select_expression = ", ".join(keys)
        query = self.getTestRunQuery(**kwargs)
        values = self.store.execute(
            "SELECT %s FROM testrun WHERE %s"
            % (select_expression, query)).get_all()
        if len(values) != 1:
            self.fail("One test run expected, %d found" % len(values))

        return dict(zip(keys, values[0]))

    def countTestRuns(self, **kwargs):
        query = self.getTestRunQuery(**kwargs)
        return self.store.execute(
            "SELECT COUNT(*) FROM testrun WHERE %s" % query).get_one()[0]

    def assertCountTestRuns(self, count, **kwargs):
        test_run_count = self.countTestRuns(**kwargs)
        self.assertEquals(test_run_count, count)


class TestTestRunInsert(TestRunTestCase):

    def test_insertTestRun(self):
        self.insertTestRun()
        self.assertCountTestRuns(1, date_created=self.default_date_created)

    def test_insertAfterTestResult(self):
        """
        Test results immediately before the test run must have the date
        deleted set to the date created of the test run.
        """
        today = self.default_date_created
        yesterday = today - timedelta(days=1)
        self.insertTestResult(date_created=yesterday)
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=today)

    def test_insertBeforeTestResult(self):
        """
        Test results immediately after the test run must remain exaclty
        the same.
        """
        today = self.default_date_created
        tomorrow = today + timedelta(days=1)
        self.insertTestResult(date_created=tomorrow)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

    def test_insertBetweenTestResults(self):
        """
        Test results immediately before the test run must have the date
        deleted set to the date created of the test run and test results
        immediately after the test run must remain exactly the same.
        """
        today = self.default_date_created
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        self.insertTestResult(status=0, date_created=yesterday)
        self.insertTestResult(status=1, date_created=tomorrow)
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=tomorrow)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=today)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

    def test_insertAfterTestResultBeforeTestRun(self):
        """
        Test results immediately before the test run must have the date
        deleted set to the date created of the test run, another test
        result must be created after the test run and the other test
        results must remain the same.
        """
        today = self.default_date_created
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        dayafter = today + timedelta(days=2)

        self.insertTestResult(status=0, date_created=yesterday)
        self.insertTestResult(status=0, date_created=tomorrow)
        self.insertTestResult(status=1, date_created=dayafter)
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=dayafter)
        self.assertCountTestResults(
            1, date_created=dayafter, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=today)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=dayafter)
        self.assertCountTestResults(
            1, date_created=dayafter, date_deleted=INFINITY_DATETIME)

    def test_insertAfterTestRun(self):
        """
        Test results before the test run must have the date deleted
        set to the date created of the test run.
        """
        today = self.default_date_created
        daybefore = today - timedelta(days=2)
        yesterday = today - timedelta(days=1)

        self.insertTestResult(date_created=daybefore)
        self.insertTestResult(date_created=yesterday)
        self.assertCountTestResults(
            1, date_created=daybefore, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=daybefore, date_deleted=today)

    def test_insertBeforeTestRun(self):
        """
        Test results after the test run must remain exaclty the same.
        """
        today = self.default_date_created
        tomorrow = today + timedelta(days=1)
        dayafter = today + timedelta(days=2)

        self.insertTestResult(date_created=dayafter)
        self.assertCountTestResults(
            1, date_created=dayafter, date_deleted=INFINITY_DATETIME)

        self.insertTestRun(date_created=tomorrow)
        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=dayafter, date_deleted=INFINITY_DATETIME)

    def test_insertAfterTestRunBeforeTestResult(self):
        """
        Test results before the test run must have the date deleted set
        to the date created of the test run the test results immediately
        after must remain the same.
        """
        today = self.default_date_created
        daybefore = today - timedelta(days=2)
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)

        self.insertTestResult(status=0, date_created=daybefore)
        self.insertTestResult(status=0, date_created=yesterday)
        self.insertTestResult(status=1, date_created=tomorrow)
        self.assertCountTestResults(
            1, date_created=daybefore, date_deleted=tomorrow)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=daybefore, date_deleted=today)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

    def test_insertBetweenTestRuns(self):
        """
        Test results before the test run must have the date deleted set
        to the date created of the test run, another test result must
        be created after the test run and the other test results must
        remain the same.
        """
        today = self.default_date_created
        daybefore = today - timedelta(days=2)
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        dayafter = today + timedelta(days=2)

        self.insertTestResult(status=0, date_created=daybefore)
        self.insertTestResult(status=0, date_created=yesterday)
        self.insertTestResult(status=0, date_created=tomorrow)
        self.insertTestResult(status=1, date_created=dayafter)
        self.assertCountTestResults(
            1, date_created=daybefore, date_deleted=dayafter)
        self.assertCountTestResults(
            1, date_created=dayafter, date_deleted=INFINITY_DATETIME)

        self.insertTestRun()
        self.assertCountTestResults(
            1, date_created=daybefore, date_deleted=today)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=dayafter)
        self.assertCountTestResults(
            1, date_created=dayafter, date_deleted=INFINITY_DATETIME)

    def test_insertFoo(self):
        today = self.default_date_created
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        self.insertTestResult(status=0, date_created=yesterday)
        self.insertTestResult(status=0, date_created=tomorrow)
        self.insertTestRun()


class TestTestRunDelete(TestRunTestCase):

    def test_deleteTestRun(self):
        self.insertTestResult()
        self.assertCountTestRuns(1)
        self.assertCountTestResults(1)

        self.deleteTestRun()
        self.assertCountTestRuns(0)
        self.assertCountTestResults(0)

    def test_deleteMiddle(self):
        """
        Test results in the middle of a test run must remain exactly
        the same.
        """
        today = self.default_date_created
        yesterday = today - timedelta(days=1)
        self.insertTestResult(date_created=yesterday)
        self.insertTestResult()
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=INFINITY_DATETIME)

        self.deleteTestRun()
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=INFINITY_DATETIME)

    def test_deleteBeginning(self):
        """
        Test results at the beginning of a test run must have the date
        created set to the next test run.
        """
        today = self.default_date_created
        tomorrow = today + timedelta(days=1)
        self.insertTestResult()
        self.insertTestResult(date_created=tomorrow)
        self.assertCountTestResults(
            1, date_created=today, date_deleted=INFINITY_DATETIME)

        self.deleteTestRun()
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

    def test_deleteBefore(self):
        """
        Test results after the test run must remain exactly the same.
        """
        today = self.default_date_created
        tomorrow = today + timedelta(days=1)
        self.insertTestResult(date_created=tomorrow)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

        self.deleteTestRun()
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

    def test_deleteBetweenSame(self):
        """
        Test results before and after the test run must be merged if
        they are the same.
        """
        today = self.default_date_created
        yesterday = today - timedelta(days=1)
        tomorrow = today + timedelta(days=1)
        self.insertTestResult(status=0, date_created=yesterday)
        self.insertTestResult(status=1)
        self.insertTestResult(status=0, date_created=tomorrow)
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=today)
        self.assertCountTestResults(
            1, date_created=today, date_deleted=tomorrow)
        self.assertCountTestResults(
            1, date_created=tomorrow, date_deleted=INFINITY_DATETIME)

        self.deleteTestRun()
        self.assertCountTestResults(
            1, date_created=yesterday, date_deleted=INFINITY_DATETIME)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
