# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "AnsiParser",
    ]


# BS, HT, LF, CR, FF, ESC
VALID_ANSI_PATTERN = r"\x08\x09\x0A\x0C\x0D\x1B"


class AnsiResult:

    text = None

    def setText(self, text):
        self.text = text


class AnsiParser:
    """Parser for ansi text."""

    def __init__(self, stream):
        self.stream = stream

    def __str__(self):
        result = AnsiResult()
        self.run(result)
        return result.text

    def run(self, result):
        text = [""]
        row = -1
        col = 0
        escape = ""
        saved = [0, 0]

        ansi = self.stream.read()
        for ch in ansi:
            if ord(ch) == 0x1B or len(escape) > 0:
                # On ESC
                if chr(0x1B) in [escape, ch]:
                    escape = ""
                    if ch == "c":
                        text = [""]
                        row = -1
                        col = 0
                        saved = [0, 0]
                    elif ch == "D":
                        row += 1
                        if row == 0:
                            row = -1
                            text.append("")
                    elif ch == "M":
                        row -= 1
                        if row < -len(text):
                            text = [""] + text
                    elif ch == "7":
                        saved = [row + len(text), col]
                    elif ch == "8":
                        [row, col] = saved
                        row -= len(text)
                    elif ord(ch) in [0x1B, 0x5B]:
                        escape = ch
                    continue
                # Just after hitting the extended ESC marker
                elif escape == "[":
                    escape = ""

                if ch in "0123456789;":
                    escape += ch
                    continue
                elif ch in "Hf":
                    opts = escape.split(";") + ["", ""]
                    row = -len(text) + max(0, int("0" + opts[0]) - 1)
                    col = max(0, int("0" + opts[1]) - 1)
                elif ch in "s":
                    saved = [row + len(text), col]
                elif ch in "u":
                    [row, col] = saved
                    row -= len(text)
                elif ch in "K":
                    if escape == "1":
                        text[row] = " " * (col + 1) + text[row][col + 1:]
                    elif escape == "2":
                        text[row] = ""
                    else:
                        text[row] = text[row][:col]
                elif ch in "J":
                    if len(escape) == 0:
                        text = text[:row] + [""]
                    else:
                        for i in range(row + len(text) + 1):
                            text[i] = ""
                elif ch in "A":
                    row -= max(1, int("0" + escape.split(";")[0]))
                    if row <= len(text):
                        row = -len(text)
                elif ch in "B":
                    row += max(1, int("0" + escape.split(";")[0]))
                    while row >= 0:
                        text.append("")
                        row -= 1
                elif ch in "C":
                    col += max(1, int("0" + escape.split(";")[0]))
                elif ch in "D":
                    col = max(0, col - max(1, int("0" + escape.split(";")[0])))

                escape = ""
                continue

            # Control char
            if ch in "\r\n\f\t\b":
                if ch == "\r":
                    col = 0
                elif ch in "\n\f":
                    row += 1
                    if row == 0:
                        row = -1
                        text.append("")
                    col = 0
                elif ch == "\t":
                    col = (col + 8) & ~7
                elif ch == "\b":
                    col = max(0, col - 1)
                continue

            if len(text[row]) < col:
                text[row] += " " * (col - len(text[row]))
            text[row] = text[row][:col] + ch + text[row][col + 1:]
            col += 1

        result.setText("\n".join(text))
        return result
