# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "create_from_file",
    ]

try:
    import xml.etree.cElementTree as etree
except ImportError:
    import cElementTree as etree

import re

from StringIO import StringIO
from logging import getLogger
from pkg_resources import resource_string

from lpresults.registry.formatters import string_to_datetime

from lpresults import xunit
from lpresults.xunit.dispatcher import DispatcherQueue
from lpresults.xunit.parsers.ansi import AnsiParser
from lpresults.xunit.parsers.cpuinfo import CpuinfoParser
from lpresults.xunit.parsers.cputable import CputableParser
from lpresults.xunit.parsers.deferred import DeferredParser
from lpresults.xunit.parsers.dmesg import DmesgParser
from lpresults.xunit.parsers.dmidecode import DmidecodeParser
from lpresults.xunit.parsers.dmisys import DmiSysParser
from lpresults.xunit.parsers.udevadm import UdevadmParser
from lpresults.xunit.protocol import Protocol
from lpresults.xunit.status import Status
from lpresults.xunit.validators import INVALID_XML_RE


CHECKBOX_HEADER = """<?xml version="1.0" ?>"""


# Validator functions for common invalid patterns.
INVALID_TO_VALID = {
    ur"<comment>.*?</comment>":
        lambda text: str(AnsiParser(StringIO(text))),
    ur"<property .*?</property>":
        lambda text: INVALID_XML_RE.sub("", text),
    ur"\xFF{8,}":
        lambda text: "",
    }

INVALID_PATTERN = "|".join(["(%s)" % text for text in INVALID_TO_VALID.keys()])
INVALID_RE = re.compile(INVALID_PATTERN, re.M | re.S)

VALID_FUNCTIONS = INVALID_TO_VALID.values()


class CheckboxResult:

    def __init__(self, test_run_factory, **kwargs):
        self.test_run_factory = test_run_factory
        self.test_run_kwargs = kwargs
        self.dispatcher = DispatcherQueue()

        # Register handlers to incrementally add information
        register = self.dispatcher.registerHandler
        register(("cpu", "architecture",), self.addCpuArchitecture)
        register(("identifier",), self.addIdentifier)
        register(("system_unit", "device",), self.addDeviceState)
        register(("test_run", "test_result",), self.addTestResult)

        # Register handlers to set information once
        register(("architecture",), self.setArchitecture, count=1)
        register(
            ("cpuinfo", "machine", "cpuinfo_result",),
            self.setCpuinfo, count=1)
        register(
            ("project", "series",),
            self.setTestRun, count=1)
        register(
            ("system_unit", "processor",),
            self.setProcessorState, count=1)
        register(
            ("test_run", "model", "make", "version", "identifiers",),
            self.setSystemUnit, count=1)
        register(
            ("udevadm", "bits", "udevadm_result",),
            self.setUdevadm, count=1)

        # Publish events passed as keyword arguments
        if "project" in kwargs:
            self.dispatcher.publishEvent("project", kwargs.pop("project"))
            self.dispatcher.publishEvent("series", kwargs.pop("series", None))

    def addAddress(self, address):
        self.dispatcher.publishEvent("identifier", address)

    def addContext(self, text, command=None):
        if text.strip() == "Command not found.":
            return

        parsers = {
            "cat /proc/cpuinfo": self.parseCpuinfo,
            "cat /var/log/dmesg | ansi_parser": DmesgParser,
            "dmidecode": DmidecodeParser,
            "grep -r . /sys/class/dmi/id/ 2&gt;/dev/null": DmiSysParser,
            "udevadm info --export-db": self.parseUdevadm,
            }
        parser = parsers.get(command)
        if parser:
            stream = StringIO(unicode(text))
            p = parser(stream)
            p.run(self)

    def addCpu(self, cpu):
        self.dispatcher.publishEvent("cpu", cpu)

    def addCpuArchitecture(self, cpu, architecture):
        if cpu["debian_name"] == architecture:
            self.dispatcher.publishEvent("machine", cpu["gnu_name"])
            self.dispatcher.publishEvent("bits", cpu["bits"])

    def addDevice(self, device):
        self.dispatcher.publishEvent("device", device)

    def addDeviceState(self, system_unit, device):
        system_unit.addDeviceState(
            bus_name=device.bus, category_name=device.category,
            product_name=device.product, vendor_name=device.vendor,
            product_id=device.product_id, vendor_id=device.vendor_id,
            subproduct_id=device.subproduct_id,
            subvendor_id=device.subvendor_id,
            driver_name=device.driver)

    def addDmiDevice(self, device):
        if device.serial:
            self.dispatcher.publishEvent("identifier", device.serial)

        if device.category in ("BOARD", "SYSTEM") \
           and device.vendor != device.product \
           and device.product is not None:
            self.dispatcher.publishEvent("model", device.product)
            self.dispatcher.publishEvent("make", device.vendor)
            self.dispatcher.publishEvent("version", device.version)

    def addIdentifier(self, identifier):
        try:
            self.identifiers.append(identifier)
        except AttributeError:
            self.identifiers = [identifier]
            self.dispatcher.publishEvent("identifiers", self.identifiers)

    def addLSBRelease(self, name, value):
        pass

    def addQuestion(self, question):
        answer_to_status = {
            "fail": Status.FAIL,
            "no": Status.FAIL,
            "pass": Status.PASS,
            "skip": Status.UNTESTED,
            "uninitiated": Status.UNINITIATED,
            "unresolved": Status.UNRESOLVED,
            "unsupported": Status.UNSUPPORTED,
            "untested": Status.UNTESTED,
            "yes": Status.PASS,
            }

        # Required arguments
        test_result = dict(
            name=question["name"],
            status=answer_to_status[question["answer"]["value"]],
            )

        # Optional arguments
        if "comment" in question:
            test_result["output"] = question["comment"]

        test_result.update(self.test_run_kwargs)
        self.dispatcher.publishEvent("test_result", test_result)

    def addTestResult(self, test_run, test_result):
        test_run.addTestResult(**test_result)

    def addSummary(self, name, value):
        if name == "architecture":
            self.dispatcher.publishEvent("architecture", value)
        elif name == "distribution":
            self.dispatcher.publishEvent("project", value)
        elif name == "distroseries":
            self.dispatcher.publishEvent("series", value)

    def parseCpuinfo(self, cpuinfo):
        self.dispatcher.publishEvent("cpuinfo", cpuinfo)
        return DeferredParser(self.dispatcher, "cpuinfo_result")

    def parseSysfs(self, sysfs):
        return DeferredParser(self.dispatcher)

    def parseUdevadm(self, udevadm):
        self.dispatcher.publishEvent("udevadm", udevadm)
        return DeferredParser(self.dispatcher, "udevadm_result")

    def setArchitecture(self, architecture):
        string = resource_string(xunit.__name__, "cputable")
        stream = StringIO(string.decode("utf-8"))
        parser = CputableParser(stream)
        parser.run(self)

    def setCpuinfo(self, cpuinfo, machine, cpuinfo_result):
        parser = CpuinfoParser(cpuinfo, machine)
        parser.run(cpuinfo_result)

    def setProcessor(self, processor):
        self.dispatcher.publishEvent("processor", processor)

    def setProcessorState(self, system_unit, processor):
        system_unit.setProcessorState(
            platform_name=processor["platform"],
            make=processor["type"], model=processor["model"],
            model_number=processor["model_number"],
            model_version=processor["model_version"],
            model_revision=processor["model_revision"],
            cache=processor["cache"], other=processor["other"],
            bogomips=processor["bogomips"], speed=processor["speed"],
            count=processor["count"])

    def setSystemUnit(self, test_run, model, make, version, identifiers):
        test_run.setSystemUnit(
            model=model, make=make, version=version,
            identifiers=identifiers, **self.test_run_kwargs)
        self.dispatcher.publishEvent("system_unit", test_run.system_unit)

    def setTestRun(self, project, series):
        test_run = self.test_run_factory(
            project=project, series=series, **self.test_run_kwargs)
        self.dispatcher.publishEvent("test_run", test_run)

    def setUdevadm(self, udevadm, bits, udevadm_result):
        parser = UdevadmParser(udevadm, bits)
        parser.run(udevadm_result)


class CheckboxProtocol(Protocol):

    def __init__(self, *args, **kwargs):
        super(CheckboxProtocol, self).__init__(*args, **kwargs)

        self.logger = getLogger()

    def _getSubmission(self):
        def repl(match):
            for index, group in enumerate(match.groups()):
                if group is not None:
                    return VALID_FUNCTIONS[index](group)

        return INVALID_RE.sub(repl, self.file.read())

    def _getClient(self, node):
        """Return a dictionary with the name and version of the client."""
        return {
            "name": node.get("name"),
            "version": node.get("version"),
            }

    def _getProperty(self, node):
        """Return the (name, value) of a property."""
        return (node.get("name"), self._getValueAsType(node))

    def _getProperties(self, node):
        """Return a dictionary of properties."""
        properties = {}
        for child in node.getchildren():
            assert child.tag == "property", \
                "Unexpected tag <%s>, expected <property>" % child.tag
            name, value = self._getProperty(child)
            properties[name] = value

        return properties

    def _getValueAsType(self, node):
        """Return value of a node as the type attribute."""
        type_ = node.get("type")
        if type_ in ("bool",):
            value = node.text.strip()
            assert value in ("True", "False",), \
                "Unexpected boolean value '%s' in <%s>" % (value, node.tag)
            return value == "True"
        elif type_ in ("str",):
            return unicode(node.text.strip())
        elif type_ in ("int", "long",):
            return int(node.text.strip())
        elif type_ in ("float",):
            return float(node.text.strip())
        elif type_ in ("list",):
            return list(self._getValueAsType(child)
                for child in node.getchildren())
        elif type_ in ("dict",):
            return dict((child.get("name"), self._getValueAsType(child))
                for child in node.getchildren())
        else:
            raise AssertionError(
                "Unexpected type '%s' in <%s>" % (type_, node.tag))

    def _getValueAsBoolean(self, node):
        """Return the value of the attribute "value" as a boolean."""
        value = node.attrib["value"]
        assert value in ("True", "False",), \
            "Unexpected boolean value '%s' in tag <%s>" % (value, node.tag)
        return value == "True"

    def _getValueAsDatetime(self, node):
        """Return the value of the attribute "value" as a datetime."""
        return string_to_datetime(node.attrib["value"])

    def _getValueAsString(self, node):
        """Return the value of the attribute "value"."""
        return unicode(node.attrib["value"])

    def parseContext(self, result, node):
        """Parse the <context> part of a submission."""
        duplicates = set()
        for child in node.getchildren():
            assert child.tag == "info", \
                "Unexpected tag <%s>, expected <info>" % child.tag
            command = child.get("command")
            if command not in duplicates:
                duplicates.add(command)
                text = child.text
                if text is None:
                    text = ""
                result.addContext(text, command)
            else:
                self.logger.debug(
                    "Duplicate command found in tag <info>: %s" % command)

    def parseHardware(self, result, node):
        """Parse the <hardware> section of a submission."""
        parsers = {
            "dmi": DmidecodeParser,
            "processors": self.parseProcessors,
            "sysfs-attributes": result.parseSysfs,
            "udev": result.parseUdevadm,
            }

        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                if child.getchildren():
                    parser(result, child)
                else:
                    stream = StringIO(unicode(child.text))
                    p = parser(stream)
                    p.run(result)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <hardware>" % child.tag)

    def parseLSBRelease(self, result, node):
        """Parse the <lsb_release> part of a submission."""
        for name, value in self._getProperties(node).iteritems():
            result.addLSBRelease(name, value)

    def parsePackages(self, result, node):
        """Parse the <packages> part of a submission."""
        pass

    def parseProcessors(self, result, node):
        """Parse the <processors> part of a submission."""
        processors = []
        for child in node.getchildren():
            assert child.tag == "processor", \
                "Unexpected tag <%s>, expected <processor>" % child.tag

            # Convert lists to space separated strings.
            properties = self._getProperties(child)
            for key, value in properties.iteritems():
                if key in ("bogomips", "cache", "count", "speed",):
                    properties[key] = int(value)
                elif isinstance(value, list):
                    properties[key] = u" ".join(value)
            processors.append(properties)

        # Check if /proc/cpuinfo was parsed already.
        if any("platform" in processor for processor in processors):
            result.setProcessor(processors[0])
        else:
            lines = []
            for processor in processors:
                # Convert some keys with underscores to spaces instead.
                for key, value in processor.iteritems():
                    if "_" in key and key != "vendor_id":
                        key = key.replace("_", " ")

                    lines.append(u"%s: %s" % (key, value))

                lines.append(u"")

            stream = StringIO(u"\n".join(lines))
            parser = result.parseCpuinfo(stream)
            parser.run(result)

    def parseQuestions(self, result, node):
        """Parse the <questions> part of a submission."""
        for child in node.getchildren():
            assert child.tag == "question", \
                "Unexpected tag <%s>, expected <question>" % child.tag
            question = {
                "name": unicode(child.get("name")),
                "targets": [],
                }
            plugin = child.get("plugin", None)
            if plugin is not None:
                question["plugin"] = plugin

            answer_choices = []
            for sub_node in child.getchildren():
                sub_tag = sub_node.tag
                if sub_tag == "answer":
                    question["answer"] = answer = {}
                    answer["type"] = sub_node.get("type")
                    if answer["type"] == "multiple_choice":
                        question["answer_choices"] = answer_choices
                    unit = sub_node.get("unit", None)
                    if unit is not None:
                        answer["unit"] = unit
                    answer["value"] = sub_node.text.strip()

                elif sub_tag == "answer_choices":
                    for value_node in sub_node.getchildren():
                        answer_choices.append(
                            self._getValueAsType(value_node))

                elif sub_tag == "target":
                    # The Relax NG schema ensures that the attribute
                    # id exists and that it is an integer
                    target = {"id": int(sub_node.get("id"))}
                    target["drivers"] = drivers = []
                    for driver_node in sub_node.getchildren():
                        drivers.append(driver_node.text.strip())
                    question["targets"].append(target)

                elif sub_tag in ("comment", "command",):
                    text = sub_node.text
                    if text is None:
                        text = u""
                    else:
                        text = unicode(text.strip())
                    question[sub_tag] = text

                else:
                    raise AssertionError(
                        "Unexpected tag <%s> in <question>" % sub_tag)

            result.addQuestion(question)

    def parseSoftware(self, result, node):
        """Parse the <software> section of a submission."""
        parsers = {
            "lsbrelease": self.parseLSBRelease,
            "packages": self.parsePackages,
            }

        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                parser(result, child)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <software>" % child.tag)

    def parseSummary(self, result, node):
        """Parse the <summary> section of a submission."""
        parsers = {
            "architecture": self._getValueAsString,
            "client": self._getClient,
            "contactable": self._getValueAsBoolean,
            "date_created": self._getValueAsDatetime,
            "distribution": self._getValueAsString,
            "distroseries": self._getValueAsString,
            "kernel-release": self._getValueAsString,
            "live_cd": self._getValueAsBoolean,
            "private": self._getValueAsBoolean,
            "system_id": self._getValueAsString,
            }

        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                value = parser(child)
                result.addSummary(child.tag, value)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <summary>" % child.tag)

    def parseRoot(self, result, node):
        """Parse the <system> root of a submission."""
        parsers = {
            "context": self.parseContext,
            "hardware": self.parseHardware,
            "questions": self.parseQuestions,
            "software": self.parseSoftware,
            "summary": self.parseSummary,
            }

        # Iterate over the root children, "summary" first
        for child in node.getchildren():
            parser = parsers.get(child.tag)
            if parser:
                parser(result, child)
            else:
                self.logger.debug(
                    "Unsupported tag <%s> in <system>" % child.tag)

    def run(self, test_run_factory, **kwargs):
        submission = self._getSubmission()
        parser = etree.XMLParser()
        tree = etree.parse(StringIO(submission), parser=parser)
        root = tree.getroot()
        if root.tag != "system":
            raise AssertionError(
                "Unexpected tag <%s> at root, expected <system>" % root.tag)

        result = CheckboxResult(test_run_factory, **kwargs)
        self.parseRoot(result, root)

        return result


def create_from_file(file):
    line = file.readline()
    file.seek(0)

    if line.strip() == CHECKBOX_HEADER:
        return CheckboxProtocol(file)

    return None
